cdef Callable
cdef Sequence
cdef Literal

cdef NDArray,

cdef Vec
cdef Mat
cdef KSP

cdef BV
cdef DS
cdef FN
cdef RG
cdef ST
cdef EPS
cdef PEP
cdef NEP
cdef SVD
cdef MFN
cdef LME

cdef Scalar
cdef ArrayInt
cdef ArrayReal
cdef ArrayComplex
cdef ArrayScalar
cdef LayoutSizeSpec

cdef EPSStoppingFunction
cdef EPSArbitraryFunction
cdef EPSEigenvalueComparison
cdef EPSMonitorFunction
cdef PEPStoppingFunction
cdef PEPMonitorFunction
cdef NEPStoppingFunction
cdef NEPMonitorFunction
cdef SVDStoppingFunction
cdef SVDMonitorFunction
cdef MFNMonitorFunction
cdef LMEMonitorFunction
